`include "common_defs.v"

module accelerator1
  #(
    parameter WB_DWIDTH  = 32,
    parameter WB_SWIDTH  = 4
    )
   (
    input                  i_clk,
    input                  i_rst,
	
    input [31:0]           i_wb_adr,
    input [WB_SWIDTH-1:0]  i_wb_sel,
    input                  i_wb_we,
    output [WB_DWIDTH-1:0] o_wb_dat,
    input [WB_DWIDTH-1:0]  i_wb_dat,
    input                  i_wb_cyc,
    input                  i_wb_stb,
    output                 o_wb_ack,
    output                 o_wb_err
    );
	
	//Registers 
	reg [31:0] i_opt_reg = 'd0; 
	reg [31:0] acc_in_0_reg = 'd0; 
	reg [31:0] acc_in_1_reg = 'd0; 
	reg [31:0] acc_in_2_reg = 'd0; 
	reg [31:0] acc_in_3_reg = 'd0;
	reg [31:0] acc_in_4_reg = 'd0;
	reg [31:0] acc_in_5_reg = 'd0;
	reg [31:0] acc_in_6_reg = 'd0;
	reg [31:0] acc_in_7_reg = 'd0;
	reg [31:0] acc_in_8_reg = 'd0;
	reg [31:0] acc_in_9_reg = 'd0;
	reg [31:0] acc_in_10_reg = 'd0;
	reg [31:0] acc_in_11_reg = 'd0;
	reg [31:0] acc_in_12_reg = 'd0;
	reg [31:0] acc_in_13_reg = 'd0;
	reg [31:0] acc_in_14_reg = 'd0;
	reg [31:0] acc_in_15_reg = 'd0;
	reg [31:0] acc_in_16_reg = 'd0;
	
	reg [31:0] acc_out_0_reg = 'd0; 
	reg [31:0] status_reg    = 32'h00000000;
   // Wishbone interface
   reg [31:0]                                   wb_rdata32 = 'd0;
   wire                                         wb_start_write;
   wire                                         wb_start_read;
   reg                                          wb_start_read_d1 = 'd0;
   wire [31:0]                                  wb_wdata32;
   
   reg acc_start = 'd0;
   wire out_rdy;
   wire [31:0]                                  acc1_out_0;
   wire [31:0]                                  acc1_out_1;
   
   
   // ======================================
   // Instantiate accelerator core
   // ======================================
 	epsilon2Hybrid u_epsilon2 (
		.i_clk          	( i_clk             ),
		.i_rst        		( i_rst	            ),
		.i_start        	( acc_start         ),
		.o_rdy          	( out_rdy           ),
		
		.acc_in_0          	( acc_in_0_reg      ),
		.acc_in_1          	( acc_in_1_reg      ),
		.acc_in_2          	( acc_in_2_reg      ),
		.acc_in_3          	( acc_in_3_reg      ),
		.acc_in_4          	( acc_in_4_reg      ),
		.acc_in_5          	( acc_in_5_reg      ),
		.acc_in_6          	( acc_in_6_reg      ),
		.acc_in_7          	( acc_in_7_reg      ),
		.acc_in_8          	( acc_in_8_reg      ),
		.acc_in_9          	( acc_in_9_reg      ),
		.acc_in_10         	( acc_in_10_reg     ),
		.acc_in_11         	( acc_in_11_reg     ),
		.acc_in_12         	( acc_in_12_reg     ),
		.acc_in_13         	( acc_in_13_reg     ),
		.acc_in_14         	( acc_in_14_reg     ),
		.acc_in_15         	( acc_in_15_reg     ),
		.acc_in_16         	( acc_in_16_reg     ),
		
		.acc_out_0          ( acc1_out_0        )
	); 
	
	
   //=====================================================================
   //Assignments   
   // Can't start a write while a read is completing. The ack for the read cycle
   // needs to be sent first
   assign wb_start_write = i_wb_stb && i_wb_we && !wb_start_read_d1;
   assign wb_start_read  = i_wb_stb && !i_wb_we && !o_wb_ack;

   always @( posedge i_clk or posedge i_rst) begin
      if(i_rst)
        wb_start_read_d1 <= 1'b0;
      else
        wb_start_read_d1 <= wb_start_read;
   end

   assign o_wb_err = 1'd0;
   assign o_wb_ack = i_wb_stb && ( wb_start_write || wb_start_read_d1 );

   generate
      if (WB_DWIDTH == 128) 
        begin : wb128
           assign wb_wdata32   = i_wb_adr[3:2] == 2'd3 ? i_wb_dat[127:96] :
                                 i_wb_adr[3:2] == 2'd2 ? i_wb_dat[ 95:64] :
                                 i_wb_adr[3:2] == 2'd1 ? i_wb_dat[ 63:32] :
                                 i_wb_dat[ 31: 0] ;
           
           assign o_wb_dat    = {4{wb_rdata32}};
        end
      else
        begin : wb32
           assign wb_wdata32  = i_wb_dat;
           assign o_wb_dat    = wb_rdata32;
        end
   endgenerate
   // ========================================================
   // Register Writes
   // ========================================================
   	always @(posedge acc_start or posedge out_rdy)
	begin
		if(out_rdy)
		begin
			acc_out_0_reg <= acc1_out_0;
			status_reg 	  <= 32'h11111111;
		end
		
		if(acc_start)
			status_reg 	  <= 32'h00000000;
	end
	 
   always @( posedge i_clk or posedge i_rst) 
   begin
		if(i_rst)
			begin 
			i_opt_reg 		<= 32'h00000000;
			acc_in_0_reg 	<= 32'h00000000;
			acc_in_1_reg 	<= 32'h00000000;
			acc_in_2_reg 	<= 32'h00000000;
			acc_in_3_reg 	<= 32'h00000000;
			acc_in_4_reg 	<= 32'h00000000;
			acc_in_5_reg 	<= 32'h00000000;
			acc_in_6_reg 	<= 32'h00000000;
			acc_in_7_reg 	<= 32'h00000000;
			acc_in_8_reg 	<= 32'h00000000;
			acc_in_9_reg 	<= 32'h00000000;
			acc_in_10_reg 	<= 32'h00000000;
			acc_in_11_reg 	<= 32'h00000000;
			acc_in_12_reg 	<= 32'h00000000;
			acc_in_13_reg 	<= 32'h00000000;
			acc_in_14_reg 	<= 32'h00000000;
			acc_in_15_reg 	<= 32'h00000000;
			acc_in_16_reg 	<= 32'h00000000;
			acc_start 		<= 1'b0;
			end
			
		else 
		begin
				
			if(acc_start)
				begin
				acc_start 		<= 1'b0;
				i_opt_reg 		<= 32'h00000000;
				end
			 	
			if ( wb_start_write )
			begin
				case (i_wb_adr[11:0])
					`ACC1_OPTIONS: begin		
										i_opt_reg	 	 <= i_wb_dat[ 31: 0];
										acc_start        <= 1'b1;
					 end
				
					`ACC1_INPUT0:		acc_in_0_reg	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT1:		acc_in_1_reg	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT2:		acc_in_2_reg 	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT3:		acc_in_3_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT4:		acc_in_4_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT5:		acc_in_5_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT6:		acc_in_6_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT7:		acc_in_7_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT8:		acc_in_8_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT9:		acc_in_9_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT10:		acc_in_10_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT11:		acc_in_11_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT12:		acc_in_12_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT13:		acc_in_13_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT14:		acc_in_14_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT15:		acc_in_15_reg  	 <= i_wb_dat[ 31: 0];
					`ACC1_INPUT16:		acc_in_16_reg  	 <= i_wb_dat[ 31: 0];
				endcase
			end
		end
	end
   // ========================================================
   // Register Reads
   // ========================================================	  
   always @( posedge i_clk or posedge i_rst ) 
   begin
      if(i_rst) 
      	begin
        wb_rdata32 <= 32'h10101010;
      	end
      else 
      	begin
        if ( wb_start_read )
			case ( i_wb_adr[11:0] )
				`ACC1_OPTIONS: wb_rdata32 <= i_opt_reg[ 31: 0];
				`ACC1_OUTPUT0: wb_rdata32 <= acc_out_0_reg[ 31: 0];
				`ACC1_STATUS:  wb_rdata32 <= status_reg[ 31: 0];
				
				default: wb_rdata32 <= 32'h33333333;
			endcase
      	end
   end
   
endmodule
