library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
library work;
use work.fixed_point_pkg.all;

entity epsilon2FixedPoint is
port (clk 			: in std_logic;
		i_start		:  in std_logic;
		rst			:  in std_logic;
		diffpoint0  :  in signed(31 downto 0);
		diffpoint1  :  in signed(31 downto 0);
		diffpoint2  :  in signed(31 downto 0);
		diffpoint3  :  in signed(31 downto 0);
		diffpoint4  :  in signed(31 downto 0);
		diffpoint5  :  in signed(31 downto 0);
		diffpoint6  :  in signed(31 downto 0);
		xV20        :  in signed(31 downto 0);
		xV21        :  in signed(31 downto 0);
		xV22        :  in signed(31 downto 0);
		xV23        :  in signed(31 downto 0);
		xV24        :  in signed(31 downto 0);
		xV25        :  in signed(31 downto 0);
		xV26        :  in signed(31 downto 0);
		mulpoint	:  in signed(31 downto 0);
		d1			:  in signed(31 downto 0);
		df			:  in signed(31 downto 0);
		fractionIN  :  in std_logic_vector(31 downto 0);
		fractionOUT	:  out std_logic_vector(31 downto 0);
		o_rdy       :  out std_logic := '0'
    );
end epsilon2FixedPoint;

architecture Behavioral of epsilon2FixedPoint is
signal nom64      		: signed(63 downto 0) := (others => '0'); 
signal nom32      		: std_logic_vector(31 downto 0);
signal denom64    		: signed(63 downto 0) := (others => '0');
signal denom32    		: std_logic_vector(31 downto 0);
signal fraction    		: std_logic_vector(31 downto 0);
signal fractionTemp 	: signed(31 downto 0);
signal nd				: std_logic:= '0';
signal rdy				: std_logic;
signal rfd				: std_logic; 


shared variable denomTrail64	: integer range 0 to 64 :=0;
shared variable denomLead64		: integer range 0 to 64 :=0;
shared variable nomTrail64		: integer range 0 to 64 :=0;
shared variable nomLead64		: integer range 0 to 64 :=0;
shared variable nomTrail32		: integer range 0 to 32 :=0;
shared variable nomLead32		: integer range 0 to 32 :=0;
shared variable denomTrail32	: integer range 0 to 32 :=0;
shared variable denomLead32		: integer range 0 to 32 :=0;
signal ignoreFlag				: std_logic:= '0';
signal signDenom				: std_logic:= '0';
	
signal temp64	: signed(63 downto 0) := (others => '0'); --intermediate multiplication result 32x32=>64 bits
signal temp32	: signed(31 downto 0) := (others => '0'); --final result 
signal temp1   	: std_logic_vector(31 downto 0) := (others => '0');
type state is (idleS, S1, S2, S3, S4, S5, S6, S7); 
signal presentState, nextState : state;

component fixedPointDivider
	port (
	clk		: in std_logic;
	nd		: in std_logic;
	rdy		: out std_logic;
	rfd		: out std_logic;
	dividend: in std_logic_vector(31 downto 0);
	divisor	: in std_logic_vector(31 downto 0);
	quotient: out std_logic_vector(31 downto 0));
end component;


begin

  	fractionOUT <= temp1;

	divider : fixedPointDivider
		port map (
			clk 	 => clk,
			nd 		 => nd,
			rdy 	 => rdy,
			rfd 	 => rfd,
			dividend => nom32,
			divisor  => denom32,
			quotient => fraction
		);

	
--	Processes for zeros detection
	denom64Zeros: process(denom64) is
	begin
		denomLead64 :=leading64(std_logic_vector(denom64));
		denomTrail64:=trailing64(std_logic_vector(denom64));
	end process denom64Zeros;
	
	nom64Zeros: process(nom64) is
	begin
		nomLead64 	:=leading64(std_logic_vector(nom64));
		nomTrail64	:=trailing64(std_logic_vector(nom64));
	end process nom64Zeros;

	denom32Zeros: process(denom32) is
	begin
		denomLead32 :=leading32(denom32);
		denomTrail32:=trailing32(denom32);
	end process denom32Zeros;
	
	nom32Zeros: process(nom32) is
	begin
		nomLead32 	:=leading32(nom32);
		nomTrail32	:=trailing32(nom32);
	end process nom32Zeros;
	
--	Rest of the code (sequential+combinatorial)
	seq: process (rst, clk) is
	begin
		if(rst='1') then
			presentState <= idleS;
		elsif (rising_edge(clk)) then
			presentState <=nextState;
		end if;
	end process seq;
	
	comb: process(i_start, presentState, rdy, nd, rfd) is
	variable Qdenom 	 : integer := 0;
	variable Qnom   	 : integer := 0;
	variable Qfraction : integer := 0;
	begin
		case presentState is
			when idleS =>	
				nd 			<= '0';
				o_rdy 		<= '0';
				signDenom 	<= '0';
				ignoreFlag 	<= '0';
				if(i_start = '1') then 
					nextState <= S1;
				else
					nextState <= idleS;
				end if;

				
			when S1 =>
				nd 			<= '0';
				o_rdy 		<= '0';
				temp1		<= fractionIN;					
				temp64 		<=diffpoint0*xV20+diffpoint1*xV21+diffpoint2*xV22+diffpoint3*xV23+diffpoint4*xV24+diffpoint5*xV25+diffpoint6*xV26;
				if((d1*df)<x"0000000000000000") then
					denom64 	<= -d1*df;
					signDenom 	<= '1';
				else
					denom64 	<= d1*df;
					signDenom 	<= '0';					
				end if;
				Qdenom 		:= 38;
				ignoreFlag 	<= '0';
				nextState 	<= S2;
				
				
			when S2 =>
				nd 			<= '0';
				o_rdy 		<= '0';
				temp32 		<= temp64(31)&temp64(31)&temp64(31 downto 2);
				if(denomLead64+denomTrail64>=32) then
					if(denomLead64<26) then
						denom32 	<= std_logic_vector(denom64(63-denomLead64 downto 32-denomLead64));
						Qdenom 		:= 6+denomLead64;
						ignoreFlag 	<= '0';
						nextState 	<= S3;
					elsif(denomLead64>=26)and(denomTrail64>=6) then
						denom32 	<=std_logic_vector(denom64(37 downto 6));
						Qdenom 		:= 32;
						ignoreFlag 	<= '0';
						nextState 	<= S3;
					else
						ignoreFlag 	<= '1';
						nextState 	<= S7;
					end if;
				else
					ignoreFlag 	<= '1';
					nextState 	<= S7;
				end if;
				
				
			when S3 =>
				nd 			<= '0';
				o_rdy 		<= '0';
				nom64		<=(mulpoint-temp32)*(mulpoint-temp32);
				Qnom		:= 40;
				ignoreFlag  <= '0';
				nextState	<= S4;
				
				
			when S4 =>
				nd 		<= '0';
				o_rdy 	<= '0';
				if(nomLead64+nomTrail64>=32) then
					if(nomLead64<24) then
						nom32 		<=std_logic_vector(nom64(63-nomLead64 downto 32-nomLead64));
						Qnom 		:= 8+nomLead64;
						ignoreFlag 	<= '0';
						nextState 	<= S5;
					elsif(nomLead64>=24)and(nomTrail64>=8) then
						nom32 		<=std_logic_vector(nom64(39 downto 8));
						Qnom 		:= 32;
						ignoreFlag 	<= '0';
						nextState 	<= S5;
					else
						ignoreFlag 	<= '1';
						nextState 	<= S7;
					end if;
				else
					ignoreFlag 	<= '1';
					nextState 	<= S7;					
				end if;
				
				
			when S5 =>
				nd 		<= '0';
				o_rdy 	<= '0';	
				if(Qnom>Qdenom)and(nomTrail32>=Qnom-Qdenom) then
					nom32   	<= std_logic_vector(shift_right(unsigned(nom32),Qnom-Qdenom));
					Qnom 		:= Qdenom;
					ignoreFlag 	<= '0';
					nextState 	<= S6;
				elsif(Qnom>Qdenom)and(denomLead32>=Qnom-Qdenom) then
					denom32   	<= std_logic_vector(shift_left(unsigned(denom32),Qnom-Qdenom));
					Qdenom 		:= Qnom;
					ignoreFlag 	<= '0';
					nextState 	<= S6;
					
				elsif(Qdenom>Qnom)and(denomTrail32>=Qdenom-Qnom) then
					denom32   	<= std_logic_vector(shift_right(unsigned(denom32),Qdenom-Qnom));
					Qdenom 		:= Qnom;
					ignoreFlag 	<= '0';
					nextState 	<= S6;
					
				elsif(Qdenom>Qnom)and(nomLead32>=Qdenom-Qnom) then
					nom32   	<= std_logic_vector(shift_left(unsigned(nom32),Qdenom-Qnom));
					Qnom 		:= Qdenom;
					ignoreFlag 	<= '0';
					nextState 	<= S6;
				else
					ignoreFlag 	<= '1';
					nextState 	<= S7;					
				end if;
	
				
			when S6 =>
				nd 		<= '0';
				o_rdy 	<= '0';
				if(denom32>nom32) then -- no need to invert denom32, cause nom32 is always positive
					--division, keep this state until the divider is ready to receive data
					if(rfd='1') then
						if((nomLead32>0)and(denomTrail32>0)) then
							nom32   	<=std_logic_vector(shift_left(unsigned(nom32),7));
							denom32 	<=std_logic_vector(shift_right(unsigned(denom32),7));
							Qfraction 	:= 14;
							nd 			<= '1';
							ignoreFlag 	<= '0';
							nextState 	<= S7;
						else
							ignoreFlag 	<='1';
							nextState 	<= S7;
						end if;
					else
						nd 				<= '0';
						ignoreFlag 		<= '0';
						nextState 		<= S6;
					end if;
				else
					ignoreFlag 		<= '0';
					o_rdy 			<= '1';
					temp1		 	<= "00000000000000000100000000000000"; --"1" in Q14 format
					nextState 		<= idleS;	
				end if;
				
				
			when S7 =>
				nd 		<= '0';
				o_rdy 	<= '0';
				temp1	<= fractionIN;
				--wait until the divider has finished
				if(rdy='0') then
					if(ignoreFlag='0') then
						nd 		  	<= '0';
						nextState 	<= S7;
					else
						o_rdy 		<= '1';
						temp1		<= fractionIN;
						nextState 	<= idleS;
					end if;
				else
					o_rdy <= '1';
					if(ignoreFlag='0') then
						if(signDenom='0') then
							temp1	<=	fraction;
						else
							temp1	<=	std_logic_vector(unsigned (not fraction) + 1);
						end if;
					else
						temp1 	<= fractionIN;
					end if;
					nextState 	<= idleS;
				end if;
		end case;
	end process comb;	
					
end Behavioral;