library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
--------------------------------------------------------

entity epsilon2Hybrid is
port(	clk         :  in std_logic;
		rst			:  in std_logic;
		i_start		:  in std_logic;
		diffpoint0  :  in signed(31 downto 0);
		diffpoint1  :  in signed(31 downto 0);
		diffpoint2  :  in signed(31 downto 0);
		diffpoint3  :  in signed(31 downto 0);
		diffpoint4  :  in signed(31 downto 0);
		diffpoint5  :  in signed(31 downto 0);
		diffpoint6  :  in signed(31 downto 0);
		xV20        :  in signed(31 downto 0);
		xV21        :  in signed(31 downto 0);
		xV22        :  in signed(31 downto 0);
		xV23        :  in signed(31 downto 0);
		xV24        :  in signed(31 downto 0);
		xV25        :  in signed(31 downto 0);
		xV26        :  in signed(31 downto 0);
		mulpoint	:  in signed(31 downto 0);
		d1			:  in std_logic_vector(31 downto 0);
		df			:  in std_logic_vector(31 downto 0);
		o_rdy       :  out std_logic;
		fraction	:  out std_logic_vector(31 downto 0)
);
end epsilon2Hybrid;

--------------------------------------------------------

architecture behavior of epsilon2Hybrid is

signal result1  : signed(31 downto 0) := (others => '0'); 
signal result2  : std_logic_vector(31 downto 0) := (others => '0');  
signal resultTemp: signed(63 downto 0) := (others => '0'); --intermediate multiplication result 32x32=>64 bits
type state is (idleS, S1, S2, S3, S4, S5, S6, S7); 
signal presentState : state;

------------Components generated with CoreGenerator----------------------
	COMPONENT fixedTofloat 
	PORT (
		 clk            : IN STD_LOGIC;
		 a 	 			: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 operation_nd   : IN STD_LOGIC;
		 operation_rfd 	: OUT STD_LOGIC;
		 result 		: OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
		 rdy 			: OUT STD_LOGIC
	);
	END COMPONENT;
	
	
	COMPONENT floatMultiplier
	PORT (
		 a 				: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 b 				: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 operation_nd 	: IN STD_LOGIC;
		 operation_rfd 	: OUT STD_LOGIC;
		 clk 			: IN STD_LOGIC;
		 result 		: OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
		 rdy 			: OUT STD_LOGIC
	);
	END COMPONENT;
	
	COMPONENT floatComparator
	 PORT (
		 a 				: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 b 				: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 operation_nd 	: IN STD_LOGIC;
		 operation_rfd 	: OUT STD_LOGIC;
		 clk 			: IN STD_LOGIC;
		 result 		: OUT STD_LOGIC_VECTOR(0 DOWNTO 0);
		 rdy 			: OUT STD_LOGIC
	  );
	END COMPONENT;

	COMPONENT floatDivider
	  PORT (
		 a 	  			: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 b 	  			: IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 operation_nd 	: IN STD_LOGIC;
		 operation_rfd 	: OUT STD_LOGIC;
		 clk 	  		: IN STD_LOGIC;
		 result 		: OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
		 rdy 			: OUT STD_LOGIC
	  );
	END COMPONENT;

	signal nomtemp      : std_logic_vector(31 downto 0) := (others => '0');
	signal d1float 	   	: std_logic_vector(31 downto 0):= (others => '0');
	signal dffloat		: std_logic_vector(31 downto 0):= (others => '0');
	signal dftemp		: std_logic_vector(31 downto 0) := (others => '0'); 
	signal nomfloat     : std_logic_vector(31 downto 0); 
	signal denomfloat   : std_logic_vector(31 downto 0); 
	signal nomfloat2 	: std_logic_vector(31 downto 0);
	
	
	--Inputs - Outputs   fixed2floatConverter
	signal converterIN  : std_logic_vector(31 downto 0);
	signal converterOUT : std_logic_vector(31 downto 0);
	signal rdy1		  	: std_logic;
	signal nd1       	: std_logic;
	signal rfd1      	: std_logic := '0';
	
	--Inputs - Outputs   floatMultiplier	
	signal multiplierA  : std_logic_vector(31 downto 0);
	signal multiplierB 	: std_logic_vector(31 downto 0);
	signal multiplierRes: std_logic_vector(31 downto 0);
	signal rdy2		 	: std_logic;
	signal rfd2      	: std_logic;
	signal nd2       	: std_logic := '0';
	
	--Inputs - Outputs   floatComparator
	signal comparatorA  : std_logic_vector(31 downto 0):= (others => '0');
	signal comparatorB 	: std_logic_vector(31 downto 0):= (others => '0');
	signal comparatorRes: std_logic_vector(0 DOWNTO 0):= (others => '0');
	signal rdy3		 	: std_logic;
	signal rfd3      	: std_logic;
	signal nd3       	: std_logic := '0';
	
	--Inputs - Outputs   floatDivider
	signal dividerA   	: std_logic_vector(31 downto 0):= (others => '0');
	signal dividerB  	: std_logic_vector(31 downto 0):= (others => '0');
	signal dividerRes   : std_logic_vector(31 downto 0):= (others => '0');
	signal rdy4		 	: std_logic;
	signal rfd4      	: std_logic;
	signal nd4       	: std_logic := '0';

	signal temp		  	: std_logic_vector(3 downto 0)  := (others => '0'); 
	signal temp1   		: std_logic_vector(31 downto 0) := (others => '0');
	signal temp2   		: std_logic_vector(31 downto 0) := (others => '1');
	
	--Signals used in the FSM to define the state transitions 
	signal holdS1		: std_logic_vector(1 downto 0):= (others => '0');
	signal holdS2		: std_logic_vector(1 downto 0):= (others => '0');
	signal holdS3		: std_logic_vector(2 downto 0):= (others => '0');
	signal holdS4		: std_logic_vector(1 downto 0):= (others => '0');
	signal holdS5		: std_logic_vector(1 downto 0):= (others => '0');
	signal holdS6		: std_logic_vector(1 downto 0):= (others => '0');
	signal holdS7		: std_logic;
----------------------------------------------------------------------------
----------Registered signals according to the 'two-process' method----------

	 type registers is record
        converterIN    :  std_logic_vector(31 downto 0);
		converterOUT   :  std_logic_vector(31 downto 0);
		rfd1		   :  std_logic;
		nd1			   :  std_logic;
        multiplierA    :  std_logic_vector(31 downto 0);
		multiplierB    :  std_logic_vector(31 downto 0);
		multiplierRes  :  std_logic_vector(31 downto 0);
		rfd2		   :  std_logic;
		nd2			   :  std_logic;
        comparatorA    :  std_logic_vector(31 downto 0);
		comparatorB    :  std_logic_vector(31 downto 0);
		comparatorRes  :  std_logic_vector(31 downto 0);
		rfd3		   :  std_logic;
		nd3			   :  std_logic;			 
        dividerA	   :  std_logic_vector(31 downto 0);
		dividerB       :  std_logic_vector(31 downto 0);
		dividerRes 	   :  std_logic_vector(31 downto 0);
		rfd4		   :  std_logic;
		nd4			   :  std_logic;			 
		o_rdy		   :  std_logic;
		temp		   :  std_logic_vector(3 downto 0);
		resultTemp     :  signed(63 downto 0);
		dftemp		   :  std_logic_vector(31 downto 0);
		holdS2         :  std_logic_vector(1 downto 0);
		result1		   :  signed(31 downto 0);
		d1float        :  std_logic_vector(31 downto 0);
		holdS3         :  std_logic_vector(2 downto 0);
		dffloat		   :  std_logic_vector(31 downto 0);
		holdS4		   :  std_logic_vector(2 downto 0);
		holdS5		   :  std_logic_vector(1 downto 0);
		holdS6		   :  std_logic_vector(1 downto 0);
		holdS7		   :  std_logic;
		nomfloat	   :  std_logic_vector(31 downto 0);
		denomfloat	   :  std_logic_vector(31 downto 0);
		nomfloat2	   :  std_logic_vector(31 downto 0);
		fraction	   :  std_logic_vector(31 downto 0);
		presentState   :  state;
    end record;
    signal r, rin : registers;
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------	 

begin					  
 
	Converter: fixedTofloat 
	PORT MAP (
	 a 			    => converterIN,
	 operation_nd   => nd1,
	 operation_rfd  => rfd1,
	 clk 		    => clk,
	 result 		=> converterOUT,
	 rdy 			=> rdy1
	);	 
	
	floatMul : floatMultiplier
	PORT MAP (
	 a 				=> multiplierA,
	 b 				=> multiplierB,
	 operation_nd 	=> nd2,
	 operation_rfd  => rfd2,
	 clk 			=> clk,
	 result 		=> multiplierRes,
	 rdy 			=> rdy2
	);	
	 
  floatDiv : floatDivider
  PORT MAP (
     a 				=> dividerA,
     b 				=> dividerB,
	 operation_nd 	=> nd4,
	 operation_rfd  => rfd4,
     clk 			=> clk,
     result 		=> dividerRes,
	 rdy 			=> rdy4
    );
  
  floatComp : floatComparator
  PORT MAP (
     a 				=> comparatorA,
     b 				=> comparatorB,
	 operation_nd 	=> nd3,
	 operation_rfd  => rfd3,
     clk 			=> clk,
     result			=> comparatorRes,
	 rdy 			=> rdy3
    );
  
	
	-- Calculate the values (outputs, internal signals) 
	-- in two processes: sequential + combinatorial
	
	sequential : process(clk)
    begin
        if rising_edge(clk) then r <= rin; end if;
    end process;
	 
	 
	combinatorial : process(rst, i_start, rfd1, rdy1, rfd2, rdy2, rfd3, rdy3, rfd4, rdy4, r)
        variable v	: registers;
    begin
		  v := r;
		  	if(rst='1') then
				v.presentState := idleS;
			end if;
		  	if(i_start='1') then
				v.presentState := S1;
			end if;	
			
			case r.presentState is
			when idleS =>	  
				v.nd1	    := '0';
				v.nd2	    := '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				v.o_rdy  	:= '0';
				if(i_start = '1') then 
					v.presentState := S1;
				else
					v.presentState := idleS;
				end if;


			when S1 =>				
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				if(rfd1='1') then
					v.resultTemp 	:=diffpoint0*xV20+diffpoint1*xV21+diffpoint2*xV22+diffpoint3*xV23+diffpoint4*xV24+diffpoint5*xV25+diffpoint6*xV26;
					v.dftemp 		:= "00" & df(31 downto 2);
					v.converterIN 	:= d1;
					v.nd1			:= '1';
					v.presentState  := S2;
				else
					v.nd1		 	:= '0';
					v.presentState  := S1;
				end if;
				v.o_rdy  	:= '0';
				v.holdS2	:= "00";


			when S2 =>	
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				v.result1 	:= r.resultTemp(31)&r.resultTemp(31)&r.resultTemp(31 downto 2);
				if(rdy1='1') then
					v.d1float 	:= converterOUT;
					v.holdS2(0)	:='1';
				end if;
				if(rfd1='1') then
					v.converterIN := r.dftemp;
					v.nd1		  := '1';
					v.holdS2(1)	  := '1';										
				end if;					
				
				if(r.holdS2="11") then
					v.presentState 	:= S3;
				else
					v.presentState	:= S2;
				end if;
				v.holdS3	:="000";

				
			when S3 =>
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				if(rdy1='1') then
					v.dffloat 	:= r.converterOUT;
					v.holdS3(0)	:='1';
				end if;
				
				if(rfd2='1') then
					v.multiplierA := r.d1float;
					v.multiplierB := converterOUT;
					v.nd2		  := '1';
					v.holdS3(1)	  := '1';
				end if;
				if(rfd1='1') then
					v.converterIN := mulpoint-v.result1;
					v.nd1		  := '1';
					v.holdS3(2)	  := '1';
				end if;
				if(r.holdS3="111") then
					v.presentState 	:= S4;
				else
					v.presentState	:= S3;
				end if;
				v.holdS4	:="000";
					
					
			when S4 =>
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				if(rdy2='1') then
					v.denomfloat	:= multiplierRes;
					v.holdS4(0)		:= '1';
				end if;
				
				if(rdy1='1') then
					v.nomfloat2		:= converterOUT;
					v.holdS4(1)		:= '1';
				end if;
				
				if(rfd2='1') then
					v.multiplierA	:= converterOUT;
					v.multiplierB	:= converterOUT;
					v.nd2			:= '1';
					v.holdS4(2)		:= '1';
				end if;
				
				if(r.holdS4="111") then
					v.presentState 	:= S5;
				else
					v.presentState	:= S4;
				end if;
				v.holdS5	:="00";
					
					
			when S5 =>
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				if(rdy2='1') then
					v.nomfloat	:= multiplierRes;
					v.holdS5(0)	:= '1';
				end if;
				
				if(rfd3='1')and(v.holdS5(0) = '1') then
					v.comparatorA := r.denomfloat;
					v.comparatorB := v.nomfloat;
					v.nd3		  := '1';
					v.holdS5(1)	  := '1';
				end if;
					
				if(r.holdS5="11") then
					v.presentState 	:= S6;
					v.o_rdy  		:= '0';
				else
					v.presentState	:= S5;
				end if;
				v.holdS6	:="00";
					
			when S6 =>
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				if(rdy3='1') then
					if(comparatorRes="0") then
						if(rfd4='1') then
							v.dividerA := r.nomfloat;
							v.dividerB := r.denomfloat;
							v.nd4	   := '1';
							v.holdS6   := "01";
						else
							v.holdS6   := "00";
						end if;
					else
						v.holdS6	   := "11";
					end if;
				end if;
				if(r.holdS6="01") then
					v.presentState 	:= S7;
				elsif(r.holdS6="11") then
					v.fraction 		:= x"3f800000";
					v.o_rdy  		:= '1';
					v.presentState	:= idleS;
				else
					v.presentState	:= S6;
				end if;
				v.holdS7	:='0';


			when S7 =>
				v.o_rdy  	:= '0';
				v.nd1		:= '0';
				v.nd2		:= '0';
				v.nd3		:= '0';
				v.nd4		:= '0';
				if(rdy4='1') then
					v.fraction 		:= r.dividerRes;
					v.o_rdy     	:= '1';
					v.presentState	:= idleS;
					v.holdS7		:='1';
				else
					v.presentState	:= S7;
				end if;			
			end case;
			

			v.multiplierRes	:= multiplierRes;
			v.converterOUT 	:= converterOUT;
			v.dividerRes	:= dividerRes;
			nd1				<= v.nd1;
			nd2				<= v.nd2;
			nd3				<= v.nd3;
			nd4				<= v.nd4;
			converterIN		<= r.converterIN;
			multiplierA		<= r.multiplierA;
			multiplierB		<= r.multiplierB;
			comparatorA		<= r.comparatorA;
			comparatorB		<= r.comparatorB;
			dividerA		<= r.dividerA;
			dividerB		<= r.dividerB;
			rin 			<= v;
			fraction		<= r.fraction;
			o_rdy			<= r.o_rdy;	
	 end process;
end behavior;